/*==========================================================*/
/*  file name: random_sampler.c                             */
/*==========================================================*/
#include <stdlib.h>

#define  LIMIT_30  0x3fffffff
#define  LIMIT_31  0x7fffffff
#define  CARRY_31  0x40000000
#define  CARRY_32  0x80000000

static unsigned long  omega[5]; /* for m90random */

struct data_pair_s {            /* for DRWS */
   unsigned long x1;
   unsigned long x2;
   unsigned long a1;
   unsigned long a2;
   struct data_pair_s *next;
};
typedef struct data_pair_s  data_pair_t;

static long location;
static long locmax;
static long locmaxmax=-1;
static data_pair_t  random_list;
static data_pair_t  *current_ptr;

/*==========================================================*/
/*   Functions for pseudo-random generation "m90random"@@@*/


/* Initialization */
void  m90setseeds( unsigned long s0, unsigned long s1,
                   unsigned long s2, unsigned long s3, 
                   unsigned long s4 )
{
   omega[0] = s0 & LIMIT_30; omega[1] = s1 & LIMIT_30;
   omega[2] = s2 & LIMIT_30; omega[3] = s3 & LIMIT_30;
   omega[4] = s4 & LIMIT_30;
}

/* Returns the current seeds */
void  m90getseeds( unsigned long *sp0, unsigned long *sp1, 
                   unsigned long *sp2, unsigned long *sp3, 
                   unsigned long *sp4 )
{
   *sp0 = omega[0]; *sp1 = omega[1]; *sp2 = omega[2];
   *sp3 = omega[3]; *sp4 = omega[4];
}

/* Generates a random bit */
char  m90randombit(void)          
{
   static unsigned long alpha[5] = { /* Data of (sqrt(5)-1)/2 */
      0x278dde6e, 0x17f4a7c1, 0x17ce7301, 0x205cedc8, 0x0d042089
   };
   char data_byte;
   union bitarray {
      unsigned long of_32bits;
      char of_8bits[4];
   } data_bitarray;
   int j;
    
   for ( j=4; j>=1; ){
      omega[j] += alpha[j];
      if ( omega[j] & CARRY_31 ){ omega[j] &= LIMIT_30; omega[--j]++; }
      else --j;
   }
   omega[0] += alpha[0]; omega[0] &= LIMIT_30;
   data_bitarray.of_32bits = omega[0] ^ omega[1] ^ omega[2];
   data_byte =  data_bitarray.of_8bits[0] ^ data_bitarray.of_8bits[1] 
              ^ data_bitarray.of_8bits[2] ^ data_bitarray.of_8bits[3];
   data_byte ^= ( data_byte >> 4 ); 
   data_byte ^= ( data_byte >> 2 );
   data_byte ^= ( data_byte >> 1 );
   return( 1 & data_byte );
}

/* Generates a 31 bit random integer */
unsigned long m90random31(void)
{
   int j;
   unsigned long b=0;
   for ( j=0; j<30; j++ ) { b |= m90randombit(); b <<= 1; }
   b |= m90randombit();
   return b;
}

/* Generates a 31 bit random real in [0,1) */
double m90randomu(void)
{
   return (double)m90random31()/CARRY_32;
}

/*==========================================================*/
/*   Functions for dynamic random Weyl sampling "DRWS"      */

/* Initialization */
void init_drws(void)
{
   locmax = -1; location = -1; random_list.next = 0;
}

/* Finalization */
void end_drws(void)
{
   data_pair_t *previous_ptr;
   if (random_list.next != 0){
      current_ptr = random_list.next;
      previous_ptr = &random_list; 
      while (current_ptr -> next !=0){
         previous_ptr = current_ptr;
         current_ptr = current_ptr -> next;
      }
      free(current_ptr);
      previous_ptr -> next = 0;
      end_drws();
   }
}

/* Returns the locmax */
long get_locmax(void)
{
   return locmax;
}

/* Sets the locmaxmax */
void set_locmaxmax(long n)
{
   locmaxmax = n;
}

/* Sets the first location */
void set_first_location(void)
{
   location = -1; current_ptr = &random_list;
}

/* Generates a dynamic random Weyl sample (31 bit integer) */
unsigned long drws31(void)
{
   data_pair_t *p;

   location++;
   if ((locmaxmax > 0)&&(location > locmaxmax )) return m90random31();

   if ( location > locmax ){
      p = (data_pair_t *) malloc(sizeof(data_pair_t));
      if ( p == 0 ) return CARRY_32;

      current_ptr -> next = p;
      p -> x1 = m90random31(); p -> x2 = m90random31();
      p -> a1 = m90random31(); p -> a2 = m90random31();
      p -> next = 0;
      locmax++;
   }
   current_ptr = current_ptr -> next;
   current_ptr -> x2 += current_ptr -> a2;
   current_ptr -> x1 += current_ptr -> a1;
   if ( current_ptr -> x2 & CARRY_32 ) {
      current_ptr -> x2 &= LIMIT_31;
      current_ptr -> x1 ++;
   }
   return ( current_ptr -> x1 &= LIMIT_31 );
}

/* Generates a dynamic random Weyl sample in [0,1) */
double drwsu(void)
{
   unsigned long drws31copy = drws31();
   if ( drws31() == CARRY_32 ) return -1.0;
   else return (double)drws31copy/(double)CARRY_32;
}
/*================ End of random_sampler.c ====================*/